#! /usr/bin/perl

#   	    	Transform ISBN-10 <=> ISBN-13

#              by John Walker  --  January 2008

#   This program parses ISBN-10 and ISBN-13 specifications from
#   the command line (either, intermixed in any order), validates
#   them and, if valid, emits the code in both ISBN-10 and ISBN-13
#   form (an ISBN-13 which cannot be transformed into an ISBN-10
#   is identified as "Unmappable".  ISBNs may be specified with
#   any delimiter; the delimiter is preserved in the transformed
#   number.  Output is in both delimited and "canonical" form
#   without any delimiters.

#   Sample run:

#   ./isbniser.pl 0-309-09657-X 0.309.09657.X 1844135438 \
#    	9780385611015 978-0-471-64877-2 978.0.471.\
#	64877.2 979-0-471-64877-1 1844135437
#ISBN-10:        0-309-09657-X     030909657X  ISBN-13:  9780309096577     978-0-309-09657-7
#ISBN-10:        0.309.09657.X     030909657X  ISBN-13:  9780309096577     978.0.309.09657.7
#ISBN-10:           1844135438     1844135438  ISBN-13:  9781844135431         9781844135431
#ISBN-13:        9780385611015  9780385611015  ISBN-10:     0385611013            0385611013
#ISBN-13:    978-0-471-64877-2  9780471648772  ISBN-10:     0471648779         0-471-64877-9
#ISBN-13:    978.0.471.64877.2  9780471648772  ISBN-10:     0471648779         0.471.64877.9
#ISBN-13:    979-0-471-64877-1  9790471648771  ISBN-10:     Unmappable                      
#Invalid ISBN: 1844135437

    use strict;
    use warnings;

    for my $s (@ARGV) {
	my $n = validate_ISBN_13($s);
	if ($n ne '') {
    	    my ($isbn10, $esbn10);
	    if (substr($n, 2, 1) eq '9') {
	    	$isbn10 = 'Unmappable';
		$esbn10 = '';
	    } else {
	    	$isbn10 = substr($n, 3, 9);
	    	my $ckd = ISBN_10_check($isbn10);
	    	$isbn10 .= $ckd;
		$esbn10 = $s;
		$esbn10 =~ s/^9\D?7\D?[89]\D?//;
		$esbn10 =~ s/\d(\D*)$/$ckd$1/;
	    }
    	    printf("ISBN-13: %20s  %13s  ISBN-10:  %13s  %20s\n", $s, $n, $isbn10, $esbn10);
	} elsif (($n = validate_ISBN_10($s)) ne '') {
    	    my $isbn13 = "978". substr($n, 0, 9);
	    my $ckd = ISBN_13_check($isbn13);
	    $isbn13 .= $ckd;
	    my $delim = '';
	    $delim = $1 if ($s =~ m/([^\dxX])/);
	    my $esbn13 = "978$delim$s";
	    $esbn13 =~ s/[\dxX]([^\dxX]*)$/$ckd$1/;
    	    printf("ISBN-10: %20s  %13s  ISBN-13:  %13s  %20s\n", $s, $n, $isbn13, $esbn13);
	} else {
    	    print("Invalid ISBN: $s\n");
	}
    }
    
    #	Compute ISBN-13 check digit from a canonical ISBN-13.  Only
    #	the first 12 characters of if the ISBN are examined.

    sub ISBN_13_check {
    	my ($s) = @_;
	my (@digits, $sum, $d, $i, $f);

	@digits = unpack('C12', $s);
	$sum = 0;
	$f = 1;
	for ($i = 0; $i < 12; $i ++) {
	    $d = $digits[$i] - ord('0');
	    $sum += $f * $d;
	    $f = ($f + 2) & 3;
	}
	$sum %= 10;
	return 10 - $sum;
    }

    #	Test whether the argument is a valid ISBN-13.  Punctuation
    #	is ignored.  The function returns the canonical 13 digit
    #	ISBN-13 if valid and the null string otherwise.
    
    sub validate_ISBN_13 {
    	my ($s) = @_;
	my (@digits, $sum, $d, $i, $f);
	
	$s =~ tr/0-9//cd; 	# Delete any non-digits
	if (length($s) != 13) {
	    return '';	    	# All valid ISBN-13s are thirteen characters
	}
	@digits = unpack('C13', $s);
	$sum = 0;
	$f = 1;
	for ($i = 0; $i < 13; $i ++) {
	    $d = $digits[$i] - ord('0');
	    $sum += $f * $d;
	    $f = ($f + 2) & 3;
	}
	$sum %= 10;
	if ($sum != 0) {
	    return '';	    	# ISBN-13 checksum is incorrect
	}
	if ($s !~ m/^97[89]/) {
	    return '';	    	# This is a valid EAN-13, but not in Bookland
	}
	return $s;
    }
    
    #	Compute ISBN-10 check digit from a canonical ISBN-10.  Only
    #	the first 9 characters of if the ISBN are examined.

    sub ISBN_10_check {
    	my ($s) = @_;
	my (@digits, $sum, $d, $i);
	
	@digits = unpack('C9', $s);
	$sum = 0;
	for ($i = 0; $i < 9; $i ++) {
	    $d = $digits[$i] - ord('0');
	    $sum += (10 - $i) * $d;
	}
	$sum %= 11;
	$sum = 11 - $sum;
	return ($sum == 10) ? 'X' :
	    (($sum == 11) ? 0 : $sum);
    }
   
    #	Test whether the argument is a valid ISBN-10.  Punctuation
    #	is ignored and the "X" checksum may be either upper or
    #	lower case.  The function returns the canonical 10 character
    #	ISBN-10 if valid and the null string otherwise.
    
    sub validate_ISBN_10 {
    	my ($s) = @_;
	my (@digits, $sum, $d, $i);
	
	$s = uc($s);  	    	# Actually, just for "X"
	$s =~ tr/0-9A-Z//cd; 	# Delete any punctuation
	if (length($s) != 10) {
	    return '';	    	# All valid ISBNs are ten characters
	}
	if ($s =~ m/A-WYZ/) {
	    return '';	    	# "X" is the only letter permitted in an ISBN
	}
	if (substr($s, 0, 9) =~ m/X/) {
	    return '';	    	# "X" may only appear as the last, checksum, character
	}
	@digits = unpack('C10', $s);
	$sum = 0;
	for ($i = 0; $i < 10; $i ++) {
	    if ($digits[$i] == ord('X')) {
	    	$d = 10;
	    } else {
	    	$d = $digits[$i] - ord('0');
	    }
	    $sum += (10 - $i) * $d;
	}
	$sum %= 11;
	if ($sum != 0) {
	    return '';	    	# ISBN checksum is incorrect
	}
	return $s;
    }
